﻿// -*- Encoding: UTF-8 -*-
#include "FreeRTOSCommon.h"
#include "LPC11Uxx.h"
#include "I2CDriver.h"
#include "bitop.h"
#include <string.h>

// I2C周波数定義
#define I2C_FREQ   200000

#define I2CONSET_I2EN       (0x1<<6)  /* I2C Control Set Register */
#define I2CONSET_AA         (0x1<<2)
#define I2CONSET_SI         (0x1<<3)
#define I2CONSET_STO        (0x1<<4)
#define I2CONSET_STA        (0x1<<5)

#define I2CONCLR_AAC        (0x1<<2)  /* I2C Control clear Register */
#define I2CONCLR_SIC        (0x1<<3)
#define I2CONCLR_STAC       (0x1<<5)
#define I2CONCLR_I2ENC      (0x1<<6)

#define I2CDRIVER_BUFFER_SIZE  (I2CDRIVER_WRITE_DATA_MAX + 3)

static volatile uint8_t mSendIndex;
static volatile uint8_t mSendLength;
static volatile uint8_t mSendBuffer[I2CDRIVER_BUFFER_SIZE];
static volatile uint8_t mRecvIndex;
static volatile uint8_t mRecvLength;
static volatile uint8_t *mRecvBufferPointer;
static xSemaphoreHandle mI2CTransferSemaphore;

typedef enum {
  I2C_TRANS_IDLE,
  I2C_TRANS_BUSY,
  I2C_TRANS_ERROR,
} I2C_TRANS_STATE;
static volatile I2C_TRANS_STATE mTransState;

void I2C_IRQHandler(void)
{
  bool endTransactionFlag = false;
  uint8_t stat = LPC_I2C->STAT;
  signed portBASE_TYPE higherTaskWoken = pdFALSE;

  switch (stat) {
    
  case 0x08:
    // Start Condition -> データ転送
    mSendIndex = 0;
    mRecvIndex = 0;
    LPC_I2C->DAT = mSendBuffer[mSendIndex++];
    LPC_I2C->CONCLR = I2CONCLR_SIC | I2CONCLR_STAC;
    break;
    
  case 0x10:
    // Repeated Start -> データ転送
    mRecvIndex = 0;
    LPC_I2C->DAT = mSendBuffer[mSendIndex++];
    LPC_I2C->CONCLR = I2CONCLR_SIC | I2CONCLR_STAC;
    break;

  case 0x18:
  case 0x28: 
    // 1Byteデータ送信完了
    if (mSendIndex < mSendLength) {
      LPC_I2C->DAT = mSendBuffer[mSendIndex++];  // 次のByte送信
    } else {
      if (mRecvLength != 0) {
        LPC_I2C->CONSET = I2CONSET_STA;            // (Repeated)Start
      } else {
        LPC_I2C->CONSET = I2CONSET_STO;            // Stop
        mTransState = I2C_TRANS_IDLE;
        endTransactionFlag = true;
      }
    }
    LPC_I2C->CONCLR = I2CONCLR_SIC;
    break;

  case 0x40: 
    // デバイスアドレス(Read)送信後
    if ((mRecvIndex + 1) < mRecvLength) {
      LPC_I2C->CONSET = I2CONSET_AA;
    } else {
      LPC_I2C->CONCLR = I2CONCLR_AAC;
    }
    LPC_I2C->CONCLR = I2CONCLR_SIC;
    break;
 
  case 0x50:
    // 1B受信
    mRecvBufferPointer[mRecvIndex++] = LPC_I2C->DAT;
    if ((mRecvIndex + 1) < mRecvLength) {
      LPC_I2C->CONSET = I2CONSET_AA;
    } else {
      LPC_I2C->CONCLR = I2CONCLR_AAC;
    }
    LPC_I2C->CONCLR = I2CONCLR_SIC;
    break;
    
  case 0x58:
    // 最後の1B受信
    mRecvBufferPointer[mRecvIndex++] = LPC_I2C->DAT;
    LPC_I2C->CONSET = I2CONSET_STO;            // Stop
    LPC_I2C->CONCLR = I2CONCLR_SIC;
    mTransState = I2C_TRANS_IDLE;
    endTransactionFlag = true;
    break;

  case 0x20: // Ackが帰ってこない
  case 0x30: // Actが帰ってこない
  case 0x48: // Actが帰ってこない
  case 0x00: // BusError
  default:
    LPC_I2C->CONSET = I2CONSET_STO; // Stop
    LPC_I2C->CONCLR = I2CONCLR_SIC;
    mTransState = I2C_TRANS_ERROR;
    endTransactionFlag = true;
    break;
  }
  if (endTransactionFlag) {
    xSemaphoreGiveFromISR(mI2CTransferSemaphore, &higherTaskWoken);
  }
  portEND_SWITCHING_ISR(higherTaskWoken);
}

static bool I2CTransfer(void)
{
  mTransState = I2C_TRANS_BUSY;
  
  LPC_I2C->CONSET = I2CONSET_STA;
  xSemaphoreTake(mI2CTransferSemaphore, 1000 / portTICK_RATE_MS);
  LPC_I2C->CONCLR = I2CONCLR_STAC;

  if (mTransState == I2C_TRANS_IDLE) {
    return true;
  }
  return false;
}

bool I2CRegRead(uint8_t devAddr, uint8_t regAddr,
                void* readData, uint8_t readSize)
{
  uint8_t shiftedAddr = devAddr << 1;
  mSendBuffer[0] = shiftedAddr;
  mSendBuffer[1] = regAddr;
  mSendBuffer[2] = shiftedAddr | 0x01;
  mSendLength = 2;
  mRecvBufferPointer = readData;
  mRecvLength = readSize;
  return I2CTransfer();
}

bool I2CRegWrite(uint8_t devAddr, uint8_t regAddr,
                 const void *writeData, uint8_t writeSize)
{
  uint8_t shiftedAddr = devAddr << 1;
  mSendBuffer[0] = shiftedAddr;
  mSendBuffer[1] = regAddr;
  memcpy((uint8_t*)mSendBuffer + 2, writeData, writeSize);
  mSendLength = 2 + writeSize;
  mRecvLength = 0;
  return I2CTransfer();
}

bool I2CRegWriteByte(uint8_t devAddr, uint8_t regAddr, uint8_t data)
{
  return I2CRegWrite(devAddr, regAddr, &data, 1);
}

void InitializeI2C(void)
{
  uint32_t sclValue;
  
  // Reset解除 & Clock有効化
  LPC_SYSCON->PRESETCTRL |= (0x1<<1);
  LPC_SYSCON->SYSAHBCLKCTRL |= (1<<5);

  // IO設定
  CLEAR_SET_BIT32(LPC_IOCON->PIO0_4, 0x07, 0x01); // SCL
  CLEAR_SET_BIT32(LPC_IOCON->PIO0_5, 0x07, 0x01); // SDA
  
  // SCLL,SCLHの合計値 = SystemCoreClock / I2C_FREQ
  sclValue = SystemCoreClock / I2C_FREQ / 2;
  LPC_I2C->SCLL = sclValue;
  LPC_I2C->SCLH = sclValue;
  
  vSemaphoreCreateBinary(mI2CTransferSemaphore);
  xSemaphoreTake(mI2CTransferSemaphore, 0);

  NVIC_EnableIRQ(I2C_IRQn);
  NVIC_SetPriority(I2C_IRQn, 0);
  LPC_I2C->CONSET = I2CONSET_I2EN;
}

