﻿// -*- Encoding: UTF-8 -*-
#include "FreeRTOSCommon.h"
#include "PitchMeasure.h"
#include "Accelerometer.h"
#include "mymath.h"
#include "ff.h"

#define STEP_ACCEL_THRESHOLD                30      // この値以下の値変動は無視
#define ACCEL_MEASURE_INTERVAL_MS           80      // 80msec間隔でサンプリング
#define STEP_MIN_INTERVAL_MS                300     // 歩の進みの間隔は300msec以上
#define UPPER_FLAG_CLEAR_INTERVAL_MS        1500    // 歩の進みチェックするフラグを落とす間隔
#define MAX_STEP_INTERVAL_THRESHOLD         1500    // 1.5sec以上歩の進みがない場合は、歩いてない

#define FILTER_TAP_COUNT      5
#define FILTER_DIV            256
const static uint16_t mFilterFactor[FILTER_TAP_COUNT] = {
  2, 41, 100, 41, 2
};

#define MIN_MAX_SAMPLE_COUNT  32

static uint32_t mStepCounter;
static uint32_t mPrevStepTick;
static xQueueHandle mEventQueue;

static uint16_t FilterData(uint16_t newValue)
{
  int i;
  uint32_t sum;
  static uint16_t sample[FILTER_TAP_COUNT -1];
  
  sum = sample[FILTER_TAP_COUNT - 2] * mFilterFactor[FILTER_TAP_COUNT - 1];
  for (i = (FILTER_TAP_COUNT - 3); i >= 0; i--) {
    sum += sample[i] * mFilterFactor[i + 1];
    sample[i + 1] = sample[i];
  }
  sum += newValue * mFilterFactor[0];
  sample[0] = newValue;
  
  return (uint16_t)(sum / FILTER_DIV);
}

static void UpdateMinMax(uint16_t newValue, uint16_t *minValue, uint16_t *maxValue)
{
  if (newValue < *minValue) {
    *minValue = newValue;
  }
  if (newValue > *maxValue) {
    *maxValue = newValue;
  }
}

static uint16_t UpdateThreshold(uint16_t newValue, uint16_t *upperThreshold)
{
  int i;
  uint16_t minValue;
  uint16_t maxValue;
  uint16_t threshold;
  static uint16_t sample[MIN_MAX_SAMPLE_COUNT - 1];
  
  minValue = 0xFFFF;
  maxValue = 0;

  UpdateMinMax(sample[MIN_MAX_SAMPLE_COUNT - 2], &minValue, &maxValue);
  for (i = (MIN_MAX_SAMPLE_COUNT - 3); i >= 0; i--) {
    UpdateMinMax(sample[i], &minValue, &maxValue);
    sample[i + 1] = sample[i];
  }
  UpdateMinMax(newValue, &minValue, &maxValue);
  sample[0] = newValue;

  threshold = (maxValue + minValue) / 2;
  *upperThreshold = threshold + ((maxValue - minValue) / 6);
  return threshold;
}

static void SendWalkEvent(bool newWalk)
{
  uint8_t event;
  if (newWalk) {
    event = STEP_EVENT_WALK;
  } else {
    event = STEP_EVENT_STOP;
  }
  xQueueSend(mEventQueue, &event, 0);
}

static void PitchMeasureMain(void* param)
{
  bool prevWalk;
  bool newWalk;
  
  int resultant;
  
  uint16_t filterdData;
  uint16_t threshold;
  uint16_t upperThreshold;
  bool upperFlag;
  uint32_t upperFlagTick;

  uint16_t prevData;
  uint32_t currentTick;

  int delayTime;

  prevWalk = false;
  
  upperFlag = false;
  upperFlagTick = 0;
  prevData = 0;
  
  while (1) {

    newWalk = IsWalking();
    if (newWalk != prevWalk) {
      SendWalkEvent(newWalk);
      prevWalk = newWalk;
    }
    
    resultant = GetCurrent3AxisResultant();
    if (resultant < 0) {
      continue;
    }
    currentTick = xTaskGetTickCount();

    filterdData = FilterData((uint16_t)resultant);
    threshold = UpdateThreshold(filterdData, &upperThreshold);
    
    if (filterdData > upperThreshold) {
      upperFlag = true;
      upperFlagTick = currentTick;
    }
    if ((currentTick - upperFlagTick) > UPPER_FLAG_CLEAR_INTERVAL_MS) {
      upperFlag = false;
    }
    
    if (DIFF(prevData, filterdData) > STEP_ACCEL_THRESHOLD) {
      
      if ((prevData >= threshold)
          && (filterdData <= threshold)
          && ((currentTick - mPrevStepTick) > STEP_MIN_INTERVAL_MS)
          && upperFlag) {
        
        mStepCounter++;
        mPrevStepTick = currentTick;
        upperFlag = false;
      }
      
      prevData = filterdData;
    }
    
    delayTime = (int)currentTick + ACCEL_MEASURE_INTERVAL_MS - xTaskGetTickCount();
    if (delayTime < 0) {
      delayTime = 0;
    }
    vTaskDelay((uint32_t)delayTime);
  }
}

uint32_t GetStepCounter(void)
{
  return mStepCounter;
}

void ClearStepCounter(void)
{
  mStepCounter = 0;
}

bool IsWalking(void)
{
  uint32_t diffMs;
  diffMs = xTaskGetTickCount() - mPrevStepTick;
  if (diffMs > MAX_STEP_INTERVAL_THRESHOLD) {
    return false;
  }
  return true;
}

void InitializePitchMeasure(xQueueHandle queue)
{
  mEventQueue = queue;
  InitializeAcceleroMeter();
  xTaskCreate(PitchMeasureMain,
              (const signed char*)"PIT",
              configMINIMAL_STACK_SIZE,
              NULL,
              APP_HIGH_PRIORITY,
              NULL);  
}
