﻿// -*- Encoding: UTF-8 -*-
#include "FreeRTOSCommon.h"
#include "GPS.h"
#include "ff.h"
#include "DestinationMgr.h"

#define TIMESTAMP_FLAG_MASK 1
#define TIMESTAMP_FLAG_ABS  0
#define TIMESTAMP_FLAG_REL  1
#define MAX_VIA_POINT   15

typedef struct {
  double lat;
  double lng;
  double distanceToFinal;
} VIA_POINT;

typedef struct {
  uint8_t flag;
  uint8_t viaCount;
  uint8_t reserved[2];
  uint32_t timestamp;
  VIA_POINT viaPoint[MAX_VIA_POINT];
} DESTINAION_INFO;

#define ARRIVED_THRESHOLD    0.015    // 目的地まで15メートル以下で到着とみなす
static DESTINAION_INFO mInfo;
static uint8_t mNextViaIndex;

int GetFinalScheduledArrivalTime(void)
{
  return mInfo.timestamp;
}

void GetFinalDestinationLatLng(double *lat, double *lng)
{
  const VIA_POINT *p = &(mInfo.viaPoint[mInfo.viaCount - 1]);
  *lat = p->lat;
  *lng = p->lng;
}

double GetFinalDestinationDistance(double lat, double lng)
{
  double dstLat, dstLng;
  GetFinalDestinationLatLng(&dstLat, &dstLng);
  return CalcGPSDistance(lat, lng, dstLat, dstLng);
}

bool IsArrived(double lat, double lng)
{
  double distance = GetFinalDestinationDistance(lat, lng);
  if (distance < ARRIVED_THRESHOLD) {
    return true;
  }
  return false;
}

void GetNextDestinationLatLng(double *lat, double *lng)
{
  const VIA_POINT *p = &(mInfo.viaPoint[mNextViaIndex]);
  *lat = p->lat;
  *lng = p->lng;
}

double GetCurrentToNextDestinationDistance(double lat, double lng)
{
  double nextLat, nextLng;
  GetNextDestinationLatLng(&nextLat, &nextLng);
  return CalcGPSDistance(lat, lng, nextLat, nextLng);
}

double GetNextToLastDestinationDistance(void)
{
  const VIA_POINT *p = &(mInfo.viaPoint[mNextViaIndex]);
  return p->distanceToFinal;
}

bool UpdateNextDestination(double lat, double lng)
{
  // 前回に比べ該当ポイントまでの距離が増加した &
  // 増加開始時点からの総量が閾値を超えた場合に次のポイントに遷移する
  static double prevDistance = 100000;
  const VIA_POINT *p;
  double newDistance;
  
  if (mNextViaIndex == (mInfo.viaCount - 1)) {
    return false;
  }
  
  p = &(mInfo.viaPoint[mNextViaIndex]);
  newDistance = CalcGPSDistance(lat, lng, p->lat, p->lng);
  if (newDistance < prevDistance) {
    prevDistance = newDistance;
  } else {
    if (newDistance - prevDistance > (ARRIVED_THRESHOLD / 2)) {
      mNextViaIndex++;
      prevDistance = 100000;
      return true;
    }
  }
  return false;
}

bool LoadDestinationInfo(int timestamp)
{
  FRESULT fr;
  UINT br;
  bool result = false;
  FIL *fp = gpFil;
  fr = f_open(fp, "justdst.dat", FA_READ | FA_OPEN_EXISTING);
  if (fr != FR_OK) {
    return false;
  }
  fr = f_read(fp, &mInfo, sizeof(mInfo), &br);
  if (fr == FR_OK) {
    if ((mInfo.flag & TIMESTAMP_FLAG_MASK) == TIMESTAMP_FLAG_REL) {
      mInfo.timestamp += timestamp;
    }
    mNextViaIndex = 1;
    result = true;
  }
  f_close(fp);
  
  return result;
}

