﻿// -*- Encoding: UTF-8 -*-
#include "FreeRTOSCommon.h"
#include "uart.h"
#include "GPS.h"
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <math.h>

#define PI      3.141592653589793
#define EARTH_R 6378.137

#define GPS_BAUD_RATE_DEFAULT           9600
#define GPS_BAUD_RATE_HI                57600
#define GPS_PMTK_COMMAND_TIMEOUT_TICK   ((30 * 1000) / portTICK_RATE_MS)   // 30秒
#define GPS_PMTK_WAIT_RECV_TIMEOUT      (( 1 * 1000) / portTICK_RATE_MS)   //  1秒

static bool mValid;
static double mLat;
static double mLng;
static float mSpeed;
static int mTime;
static int mDate;
static xQueueHandle mEventQueue;
static uint8_t mUpdateIntervalSec;
static uint8_t mUpdateCounter;

/**
  GPSにPMTKコマンドを送出
  command : "PMTK～" の文字列(データ先頭$と末尾の*は含めない)
**/
static void SendPmtkCommand(const char* command)
{
  char c;
  char checkSumStr[3];
  uint8_t checkSum = 0;

  DEBUG_MSG(("[GPS] PktSend %s\n", command));
  
  SendUartChar('$');
  while ((c = *command++) != '\0') {
    checkSum ^= (uint8_t)c;
    SendUartChar(c);
  }
  SendUartChar('*');
  sprintf(checkSumStr, "%02X", checkSum);
  SendUartStr(checkSumStr);
  SendUartStr("\r\n");
}

/**
  GPSにPMTKコマンドを出しレスポンス($PMTK～)を受け取る
  command : "PMTK～" の文字列(データ先頭$と末尾の*は含めない)
  recvBuf    : 受信した "PMTK～" を格納するバッファ
  recvBufSize: recvBufのサイズ

  true  : 受信成功
  false : 受信できないタイムアウト
**/
static bool DoPmtkCommand(const char* command, char* recvBuf, int recvBufSize)
{
  const char* recvLine;
  uint32_t startTick;

  SendPmtkCommand(command);
  startTick = xTaskGetTickCount();
  do {
    recvLine = AcquireUartRecvLineBuffer(GPS_PMTK_WAIT_RECV_TIMEOUT);
    if (recvLine != NULL) {
      if (strncmp(recvLine, "$PMTK", 5) == 0) {
        DEBUG_MSG(("[GPS] PktRecv %s\n", recvLine));
        if (recvBuf != NULL) {
          strncpy(recvBuf, recvLine, recvBufSize - 1);
          recvBuf[recvBufSize - 1] = '\0';
        }
        return true;
      }
      ReleaseUartRecvLineBuffer();
    }
  } while ((xTaskGetTickCount() - startTick) < GPS_PMTK_COMMAND_TIMEOUT_TICK);
  DEBUG_MSG(("[GPS] PktTimeout\n"));
  return false;
}

/**
  GPS baudレート変更
**/
static bool ChangeGPSBaud(int baud)
{
  char command[20];
  const char* recvLine;
  
  sprintf(command, "PMTK251,%d", baud);
  SendPmtkCommand(command);
  vTaskDelay(1000);
  
  SetUartBaudRate(baud);
  
  recvLine = AcquireUartRecvLineBuffer(GPS_PMTK_COMMAND_TIMEOUT_TICK);
  if  (recvLine == NULL) {
    return false;
  }
  DEBUG_MSG(("[GPS] ChangeRecv %s\n", recvLine));
  ReleaseUartRecvLineBuffer();

  recvLine = AcquireUartRecvLineBuffer(GPS_PMTK_COMMAND_TIMEOUT_TICK);
  if  (recvLine == NULL) {
    return false;
  }
  DEBUG_MSG(("[GPS] ChangeRecv2 %s\n", recvLine));
  ReleaseUartRecvLineBuffer();
  
  return DoPmtkCommand("PMTK000", NULL, 0);
}

/**
  dddmm.mmmm 表記を ddd.dddd度 
**/
static double ConvertLatLng(double s)
{
  int d = ((int)s) / 100;
  double m = s - (d * 100);
  return (d + m / 60);
}

static void ParseGPRMC(const char* gprmc)
{
  float ftime;
  char status, ns, we;
  double lat, lng;
  int date;
  int itime;
  int result;
  
  result = sscanf(gprmc,
                  "$GPRMC,%f,%c,%lf,%c,%lf,%c,%*f,%*f,%d",
                  &ftime,
                  &status,
                  &lat,
                  &ns,
                  &lng,
                  &we,
                  &date);
  if (result != 7) {
    return;
  }
  if (status != 'A') {
    mValid = false;
    return;
  }

  // 緯度/経度を百分率表記に直す
  lat = ConvertLatLng(lat);
  if (ns == 'S') {
    lat = -lat;
  }
  lng = ConvertLatLng(lng);
  if (we == 'W') {
    lng = -lng;
  }

  itime = (int)ftime;
  taskENTER_CRITICAL();
  mLat = lat;
  mLng = lng;
  mTime = itime;
  mDate = date;
  taskEXIT_CRITICAL();
  
  mValid = true;
}

static void ParseGPVTG(const char* gpvtg)
{
  float speed;
  int result;

  if (!mValid) {
    return;
  }
  result = sscanf(gpvtg, "$GPVTG,%*f,T,,M,%*f,N,%f", &speed);
  if (result != 1) {
    return;
  }
  mSpeed = speed;
}

static void GPSMain(void *param)
{
  const char* recvLine;
  uint8_t event;
  bool prevActive, newActive;
  bool rmcParsed;
  
  InitializeUart(GPS_BAUD_RATE_DEFAULT);
  while (1) {
    recvLine = AcquireUartRecvLineBuffer(portMAX_DELAY);
    prevActive = IsGPSActive();
    rmcParsed =  false;
    if (strncmp(recvLine, "$GPRMC", 6) == 0) {
      ParseGPRMC(recvLine);
      rmcParsed = true;
    } else if (strncmp(recvLine, "$GPVTG", 6) == 0) {
      ParseGPVTG(recvLine);
    }
    
    newActive = IsGPSActive();
    if (newActive != prevActive) {
      if (newActive) {
        event = GPS_EVENT_ACTIVE;
      } else {
        event = GPS_EVENT_INACTIVE;
      }
      xQueueSend(mEventQueue, &event, 0);
    }
    if (newActive && rmcParsed) {
            
      if (mUpdateIntervalSec != 0) {
        if (mUpdateCounter == 0) {
          event = GPS_EVENT_UPDATE;
          xQueueSend(mEventQueue, &event, 0);
          mUpdateCounter = mUpdateIntervalSec - 1;
        } else {
          mUpdateCounter--;          
        }
      }
    }
    ReleaseUartRecvLineBuffer();
  }
}

static void ConvertIntTimeToGPSTime(int itime, GPS_TIME *tm)
{
  tm->sec = itime % 100;
  itime /= 100;
  tm->min = itime % 100;
  tm->hour = itime / 100;
}

static void ConvertDateToGPSTime(int date, GPS_TIME *tm)
{
  tm->yearBase2000 = date % 100;
  date /= 100;
  tm->mon = date % 100;
  tm->day = date / 100; 
}

bool IsGPSActive(void)
{
  return mValid;
}

void GetGPSLatLng(double *lat, double *lng)
{
  taskENTER_CRITICAL();
  *lat = mLat;
  *lng = mLng;
  taskEXIT_CRITICAL();
}

float GetGPSSpeed(void)
{
  return mSpeed;
}

void GetGPSTime(GPS_TIME* tm)
{
  int itime, date;
  
  taskENTER_CRITICAL();
  itime = mTime;
  date = mDate;
  taskEXIT_CRITICAL();

  ConvertIntTimeToGPSTime(itime, tm);
  ConvertDateToGPSTime(date, tm);
}

int GetGPSTimestamp(void)
{
  // 一日内の時刻(秒)を返す
  GPS_TIME tm;
  ConvertIntTimeToGPSTime(mTime, &tm);
  return (tm.hour * 3600) + (tm.min * 60) + tm.sec;
}

int DiffGPSTimestamp(int oldTimestamp, int newTimestamp)
{
  // GetGPSTimestamp()で取得した値の差分を計算
  // 差分が一日以上あく事がないことが前提
  if (newTimestamp > oldTimestamp) {
    return newTimestamp - oldTimestamp;
  } else {
    return ((24 * 3600) - oldTimestamp) + newTimestamp;
  }
}

double CalcGPSDistance(double latA, double lngA,
                       double latB, double lngB)
{
  double latRad, lngRad, latKm, lngKm;
  
  latRad = PI / 180 * fabs(latA - latB);
  lngRad = PI / 180 * fabs(lngA - lngB);

  latKm = EARTH_R * latRad;
  lngKm = cos(PI / 180 * latA) * EARTH_R * lngRad;
  
  return sqrt(pow(latKm, 2) + pow(lngKm, 2));
}

void ChangeGPSUpdateInterval(uint8_t sec)
{
  mUpdateCounter = sec;
  mUpdateIntervalSec = sec;
}

void InitializeGPS(xQueueHandle queue)
{
  mEventQueue = queue;
  xTaskCreate(GPSMain, (const signed char*)"GPS", 150, NULL, APP_NORMAL_PRIORITY, NULL);
}
