﻿// -*- Encoding: UTF-8 -*-
#include "LPC11Uxx.h"
#include "FreeRTOSCommon.h"
#include "uart.h"
#include "bitop.h"
#include "ConstMem.h"

#define TXD_PIN PIO0_19
#define TXD_PIN_FUNC_MASK       0x07
#define TXD_PIN_FUNC_GPIO       0x00
#define TXD_PIN_FUNC_TXD        0x01
#define RXD_PIN PIO0_18
#define RXD_PIN_FUNC_MASK       0x07
#define RXD_PIN_FUNC_GPIO       0x00
#define RXD_PIN_FUNC_RXD        0x01

#define UART_RECV_LINE_BUF_SIZE         (150)
#define UART_RECV_LINE_BUF_COUNT         2

static xQueueHandle mRecvLineQueue;
static char *mBuffer[UART_RECV_LINE_BUF_COUNT];
static unsigned char mBufferIndex;
static int mRecvSize;

void UART_IRQHandler(void)
{
  uint8_t IIRVal;
  signed portBASE_TYPE higherTaskWoken = pdFALSE;

  IIRVal = LPC_USART->IIR;
  if (((IIRVal >> 1) & 0x07) == 0x02) { // IIR.INTID = 2 (Receive Data Available)
    if (LPC_USART->LSR & (1 << 0)) { // LSR.RDR
      
      char data = LPC_USART->RBR;
      if ((data == '\n') || (data == '\r')) {
        mBuffer[mBufferIndex][mRecvSize] = '\0';
        if (xQueueSendToBackFromISR(mRecvLineQueue, &mBufferIndex, &higherTaskWoken)) {
          if (mBufferIndex == (UART_RECV_LINE_BUF_COUNT - 1)) {
            mBufferIndex = 0;
          } else {
            mBufferIndex++;
          }
        }
        mRecvSize = 0;
      } else if (mRecvSize < (UART_RECV_LINE_BUF_SIZE - 1)) {
        mBuffer[mBufferIndex][mRecvSize] = data;
        mRecvSize++;
      }
    }
  } 
  portEND_SWITCHING_ISR(higherTaskWoken);
}

const char* AcquireUartRecvLineBuffer(portTickType ticksToWait)
{
  unsigned char bufferIndex;
  if (xQueuePeek(mRecvLineQueue, &bufferIndex, ticksToWait)) {
    return mBuffer[bufferIndex];
  }
  return NULL;
}

void ReleaseUartRecvLineBuffer(void)
{
  unsigned char bufferIndex;
  xQueueReceive(mRecvLineQueue, &bufferIndex, portMAX_DELAY);
}

void SendUartChar(char c)
{
  while (!(LPC_USART->LSR & (1 << 5))); // wait for LSR.THRE = 1(empty)
  LPC_USART->THR = c;
}

void SendUartStr(const char* str)
{
  char c;
  while ((c = *str++) != '\0') {
    SendUartChar(c);
  }
}

void SetUartBaudRate(int baud)
{
  uint32_t mainClk, pclk, div;
  uint8_t dlm, dll;

  mainClk = SystemCoreClock * LPC_SYSCON->SYSAHBCLKDIV;
  pclk = mainClk / LPC_SYSCON->UARTCLKDIV;
  div = pclk / (16 * baud);
  dlm = div / 256;
  dll = div % 256;

  LPC_USART->IER = 0;
  NVIC_DisableIRQ(UART_IRQn);
  
  LPC_USART->LCR |= 0x80; // LCR.DLAB = 1
  LPC_USART->DLM = dlm;
  LPC_USART->DLL = dll;
  LPC_USART->LCR &= ~0x80; // LCR.DLAB = 0
  LPC_USART->FCR = 0x07;   // FIFO Reset

  mBufferIndex = 0;
  mRecvSize = 0;
  xQueueReset(mRecvLineQueue);

  NVIC_SetPriority(UART_IRQn, 1);
  NVIC_EnableIRQ(UART_IRQn);
  LPC_USART->IER = (1 << 0); // Enables the Receive Data Available Interrupt
}

void InitializeUart(int baud)
{
  int i;
  
  mRecvLineQueue = xQueueCreate(UART_RECV_LINE_BUF_COUNT - 1, sizeof(unsigned char));

  for (i = 0; i < UART_RECV_LINE_BUF_COUNT; i++) {
    mBuffer[i] = pvPortMalloc(UART_RECV_LINE_BUF_SIZE);
  }
  
  NVIC_DisableIRQ(UART_IRQn);
  
  // Enable UART Clock
  LPC_SYSCON->SYSAHBCLKCTRL |= (1 << 12);

  // PCLK_UART分周なし(PCLK = 48MHz)
  // Fractional Dividerは使わない(LPC_USART->FDRはデフォルト値のまま)
  LPC_SYSCON->UARTCLKDIV = 1;

  // 8bit / non parity / 1bit stop
  LPC_USART->LCR = 0x03;

  // Pin設定
  CLEAR_SET_BIT32(LPC_IOCON->TXD_PIN, TXD_PIN_FUNC_MASK, TXD_PIN_FUNC_TXD);
  CLEAR_SET_BIT32(LPC_IOCON->RXD_PIN, RXD_PIN_FUNC_MASK, RXD_PIN_FUNC_RXD);
  
  SetUartBaudRate(baud);
}

