﻿// -*- Encoding: UTF-8 -*-
#include "FreeRTOSCommon.h"
#include "LPC11Uxx.h"
#include "bitop.h"

#define LED_REQ_ON         0
#define LED_REQ_OFF        1
#define LED_REQ_ON_ONCE    2
#define LED_REQ_BLINK      3
#define MAKE_LED_REQ(command, param)   ((command) | ((param) << 8))
#define EXTRACT_LED_REQ_COMMAND(req)   ((req) & 0xFF)
#define EXTRACT_LED_REQ_PARAM(req)     ((req) >> 8)

typedef enum {
  WAIT_REQUEST,
  ON_ONCE, 
  BLINK_TO_OFF,
  BLINK_TO_ON
} LED_MAIN_STATE;

#define LED_QUEUE_LEN 2
static xQueueHandle mLedQueue;

static void TurnOnLed(void)
{
  LPC_GPIO->SET[0] = (1 << 7);
}

static void TurnOffLed(void)
{
  LPC_GPIO->CLR[0] = (1 << 7);
}

static void LEDMain(void *param)
{
  uint32_t request;
  portTickType nextWaitTick = portMAX_DELAY;
  LED_MAIN_STATE state = WAIT_REQUEST;
  while (1) {
    if (xQueueReceive(mLedQueue, &request, nextWaitTick)) {
      switch (EXTRACT_LED_REQ_COMMAND(request)) {
      case LED_REQ_ON:
        TurnOnLed();
        nextWaitTick = portMAX_DELAY;
        state = WAIT_REQUEST;
        break;
      case LED_REQ_OFF:
        TurnOffLed();
        nextWaitTick = portMAX_DELAY;
        state = WAIT_REQUEST;
        break;
      case LED_REQ_ON_ONCE:
        TurnOnLed();
        nextWaitTick = EXTRACT_LED_REQ_PARAM(request);
        state = ON_ONCE;
        break;
      case LED_REQ_BLINK:
        TurnOnLed();
        nextWaitTick = EXTRACT_LED_REQ_PARAM(request);
        state = BLINK_TO_OFF;
        break;
      default: // ありえないはず
        break;
      }
    } else {
      switch (state) {
      case ON_ONCE:
        TurnOffLed();
        nextWaitTick = portMAX_DELAY;
        state = WAIT_REQUEST;
        break;
      case BLINK_TO_OFF:
        TurnOffLed();
        state = BLINK_TO_ON;
        break;
      case BLINK_TO_ON:
        TurnOnLed();
        state = BLINK_TO_OFF;
        break;
      default: // ありえないはず・・・
        TurnOffLed();
        nextWaitTick = portMAX_DELAY;
        state = WAIT_REQUEST;
        break;
      }
    }
  }
}

void LedOn(void)
{
  uint32_t request = LED_REQ_ON;
  xQueueSend(mLedQueue, &request, portMAX_DELAY);
}

void LedOff(void)
{
  uint32_t request = LED_REQ_OFF;
  xQueueSend(mLedQueue, &request, portMAX_DELAY);
}

void LedOnOnce(int onTimeMs)
{
  uint32_t request = MAKE_LED_REQ(LED_REQ_ON_ONCE, onTimeMs);
  xQueueSend(mLedQueue, &request, portMAX_DELAY);
}

void LedBlink(int intervalMs)
{
  uint32_t request = MAKE_LED_REQ(LED_REQ_BLINK, intervalMs);
  xQueueSend(mLedQueue, &request, portMAX_DELAY);
}

void InitializeLED(void)
{
  // PIN設定
  CLEAR_SET_BIT32(LPC_IOCON->PIO0_7, 0x07, 0x00); // GPIO
  LPC_GPIO->DIR[0] |= (1 << 7);
  TurnOffLed();

  mLedQueue = xQueueCreate(LED_QUEUE_LEN, sizeof(uint32_t));
  xTaskCreate(LEDMain,
              (const signed char*)"LED",
              configMINIMAL_STACK_SIZE,
              NULL,
              APP_NORMAL_PRIORITY,
              NULL);
}

void LedOnWithoutOS(void)
{
  CLEAR_SET_BIT32(LPC_IOCON->PIO0_7, 0x07, 0x00); // GPIO
  LPC_GPIO->DIR[0] |= (1 << 7);
  TurnOnLed();
}
