﻿// -*- Encoding: UTF-8 -*-
#include "FreeRTOSCommon.h"
#include "LPC11Uxx.h"
#include "ff.h"
#include "ConstMem.h"
#include "bitop.h"
#include <string.h>

static volatile const uint8_t *mBuffer;
static volatile uint32_t mEndIndex;
static volatile uint32_t mCurrentIndex;
static volatile uint32_t mBufferIndexMask;
static volatile bool     mUseSD;
static volatile uint8_t  mInterruptDivideCounter;
static xQueueHandle      mPlayingQueue;
static xQueueHandle      mEndQueue;
static xSemaphoreHandle  mMutex;
static uint8_t  *mSDBuffer;

#define QUEUE_LENGTH      4
#define EVENT_PLAY_END    0
#define EVENT_BUF_0_FREE  1
#define EVENT_BUF_1_FREE  2
#define EVENT_NOTHING     0xFF

void TIMER32_1_IRQHandler(void)
{
  // 40KHz周期で呼ばれる割り込みハンドラ
  //  => 5分周して 8KHz 単位で処理を行う
  signed portBASE_TYPE higherTaskWoken = pdFALSE;
  
  if (mInterruptDivideCounter == 0) {
    
    uint8_t event = EVENT_NOTHING;
    
    LPC_CT32B1->MR0 = 255 - mBuffer[mCurrentIndex & mBufferIndexMask];
    mCurrentIndex++;
    
    if (mCurrentIndex == mEndIndex) {
      // 再生終了(タイマー停止)
      event = EVENT_PLAY_END;
      NVIC_DisableIRQ(TIMER_32_1_IRQn);
      LPC_CT32B1->TCR = 0;      
    } else if (mUseSD) {
      if ((mCurrentIndex & (512 - 1)) == 0) {
        // 次のバッファ読み込み要求
        if (mCurrentIndex & 512) {
          event = EVENT_BUF_0_FREE;
        } else {
          event = EVENT_BUF_1_FREE;
        }
      }
    }
    if (event != EVENT_NOTHING) {
      xQueueSendToBackFromISR(mPlayingQueue, &event, &higherTaskWoken);
    }
  }
  if (mInterruptDivideCounter == 4) {
    mInterruptDivideCounter = 0;
  } else {
    mInterruptDivideCounter++;
  }

  LPC_CT32B1->IR |= (1 << 3);
  
  portEND_SWITCHING_ISR(higherTaskWoken);
}

static void StartTimer(uint32_t index)
{
  
  mCurrentIndex = index;
  mInterruptDivideCounter = 0;
  
  // タイマーのカウントアップはPCLK(48MHz)/ 4 = 12MHz
  LPC_CT32B1->PR = (4 - 1);
  // PWM設定
  LPC_CT32B1->PWMC |= (1 << 0);
  //MR3でリセットするとともに割り込みを発生させる
  // 周期 12MHz / 300clock = 40KHz)
  LPC_CT32B1->MCR |= (1 << 9) | (1 << 10);
  LPC_CT32B1->MR3 = (300 - 1);
  // パルス幅(Hi)は0が最大 / 255が最小
  LPC_CT32B1->MR0 = 255;
  // タイマスタート
  NVIC_EnableIRQ(TIMER_32_1_IRQn);
  LPC_CT32B1->TCR = 1;
}

static void ReadBlockFromSd(void *buf)
{
  UINT readSize;
  f_read(gpFil, buf, 512, &readSize);
}

static void SoundMain(void *param)
{
  NVIC_SetPriority(TIMER_32_1_IRQn, 0);
  while (1) {
    uint8_t event = EVENT_NOTHING;
    xQueueReceive(mPlayingQueue, &event, portMAX_DELAY);
    switch (event) {
    case EVENT_PLAY_END:
      if (mUseSD) {
        f_close(gpFil);
      }
      xQueueSend(mEndQueue, NULL, 0);
      break;
    case EVENT_BUF_0_FREE:
      ReadBlockFromSd(mSDBuffer + 0);
      break;
    case EVENT_BUF_1_FREE:
      ReadBlockFromSd(mSDBuffer + 512);
      break;
    default:
      // Do Nothing
      break;
    }
  }
}

static uint32_t GetWaveDataIndex(const uint8_t *data)
{
  uint32_t i;
  for (i = 32; i < (512 - 4); i++) {
    if (memcmp(data + i, "data", 4) == 0) {
      return (i + 8);
    }
  }
  return 0;
}

void PlaySoundInSD(const char* fileName, uint32_t waitForPlay)
{
  FRESULT fr;
  FILINFO fno;
  uint32_t waveDataIndex;

  if (!xSemaphoreTake(mMutex, waitForPlay)) {
    return;
  }
  
  fr = f_stat(fileName, &fno);
  if (fr != FR_OK) {
    goto Exit;
  }
  fr = f_open(gpFil, fileName, FA_READ);
  if (fr != FR_OK) {
    goto Exit;
  }
  
  mBuffer = mSDBuffer;
  mEndIndex = fno.fsize;
  mBufferIndexMask = (1024 - 1);
  mUseSD = true;
  
  // 先頭の2BlockをReadしておく
  ReadBlockFromSd(mSDBuffer + 0);
  ReadBlockFromSd(mSDBuffer + 512);

  waveDataIndex = GetWaveDataIndex(mSDBuffer);

  // 再生開始 -> 完了街
  StartTimer(waveDataIndex);
  xQueueReceive(mEndQueue, NULL, portMAX_DELAY);
  
Exit:
  xSemaphoreGive(mMutex);
}

void PlaySoundInROM(const uint8_t *data, uint32_t size, uint32_t waitForPlay)
{
  if (!xSemaphoreTake(mMutex, waitForPlay)) {
    return;
  }

  mBuffer = data;
  mEndIndex = size;
  mBufferIndexMask = 0xFFFFFFFF;
  mUseSD = false;

  // 再生開始 -> 完了待ち
  StartTimer(0);
  xQueueReceive(mEndQueue, NULL, portMAX_DELAY);

  xSemaphoreGive(mMutex);
}

void InitializeSound(void)
{
  LPC_SYSCON->SYSAHBCLKCTRL |= (1 << 10); // Timer
  
  // Pin設定
  CLEAR_SET_BIT32(LPC_IOCON->TDO_PIO0_13, 0x07, 0x03);
  
  mSDBuffer = (uint8_t*)SOUND_BUF_BASE;
  
  mPlayingQueue = xQueueCreate(QUEUE_LENGTH, 1);
  mEndQueue = xQueueCreate(1, 0);
  mMutex = xSemaphoreCreateMutex();
  
  xTaskCreate(SoundMain,
              (const signed char*)"SND",
              100,
              NULL,
              APP_NORMAL_PRIORITY,
              NULL);
}
