﻿// -*- Encoding: UTF-8 -*-
#include "LPC11Uxx.h"
#include "FreeRTOSCommon.h"
#include "mw_usbd_rom_api.h"
#include "power_api.h"
#include "bitop.h"
#include "ConstMem.h"
#include "diskio.h"
#include <string.h>

static USBD_API_T *mUsbApi;
static USBD_HANDLE_T mUsbHandle;

extern const uint8_t gUsbDeviceDescriptor[];
extern const uint8_t gUsbConfigurationDescriptor[];
extern const uint8_t gUsbStringDescriptor[];

struct INQUIRY_STR {
  char Vendor[8];
  char Product[16];
  char Revison[4];
} __attribute__((packed));
static const struct INQUIRY_STR mInquiry = {
  "d-risso",
  "just",
  "1.0"
};


void USB_IRQHandler(void)
{
  mUsbApi->hw->ISR(mUsbHandle);
}

void MSCRead(uint32_t offset, uint8_t **bufAddr, uint32_t length)
{
  uint32_t blockOffset = offset & (512 -1);
  uint8_t *workBuf = (uint8_t*)USB_WORK_BUF_BASE;
  
  if (blockOffset == 0) {
    uint32_t blockIndex = offset / 512;
    disk_read(0, workBuf, blockIndex, 1);
  }
  
  *bufAddr = workBuf + blockOffset;
}

void MSCWrite(uint32_t offset, uint8_t **bufAddr, uint32_t length)
{
  uint32_t blockOffset = offset & (512 -1);
  uint8_t *workBuf = (uint8_t*)USB_WORK_BUF_BASE;
  memcpy(workBuf + blockOffset, *bufAddr, length);
  if (((offset + length) & (512-1)) == 0) {
    uint32_t blockIndex = offset / 512;
    disk_write(0, workBuf, blockIndex, 1);
  }
}

void MSCGetWriteBuf(uint32_t offset, uint8_t **bufAddr, uint32_t length)
{
  uint32_t blockOffset = offset & (512 -1);
  uint8_t *workBuf = (uint8_t*)USB_WORK_BUF_BASE;
  *bufAddr = workBuf + blockOffset;
}

ErrorCode_t MSCVerify(uint32_t offset, uint8_t *srcBuf, uint32_t length)
{
  // Dummy
  return LPC_OK;
}

static void InitializeUsbCore(void)
{
  uint32_t requiredMemSize;
  uint32_t constMemSize;
  ErrorCode_t initResult;
  USBD_API_INIT_PARAM_T initParam = {0};
  USB_CORE_DESCS_T desc = {0};

  // USBAPI Core
  initParam.usb_reg_base = LPC_USB_BASE;
  initParam.max_num_ep = 2;
  requiredMemSize = mUsbApi->hw->GetMemSize(&initParam);
  constMemSize = USB_HW_MEM_SIZE;
  configASSERT(requiredMemSize <= constMemSize);
  
  initParam.mem_base = USB_HW_MEM_BASE;
  initParam.mem_size = constMemSize;
  
  // USBAPI Descriptor
  desc.device_desc = (uint8_t*)gUsbDeviceDescriptor;
  desc.full_speed_desc = (uint8_t*)gUsbConfigurationDescriptor;
  desc.high_speed_desc = (uint8_t*)gUsbConfigurationDescriptor;
  desc.string_desc = (uint8_t*)gUsbStringDescriptor;
  
  initResult = mUsbApi->hw->Init(&mUsbHandle, &desc, &initParam);
  configASSERT(initResult == LPC_OK);
}

static void InitializeUsbMsc(uint32_t sectorCount)
{
  uint32_t requiredSize;
  uint32_t constSize;
  ErrorCode_t initResult;
  USBD_MSC_INIT_PARAM_T initParam = {0};

  initParam.InquiryStr = (uint8_t*)&mInquiry;
  initParam.BlockCount = sectorCount;
  initParam.BlockSize = 512;
  initParam.MemorySize = sectorCount * 512;
  initParam.intf_desc = (uint8_t*)&(gUsbConfigurationDescriptor[USB_CONFIGUARTION_DESC_SIZE]);
  requiredSize = mUsbApi->msc->GetMemSize(&initParam);
  constSize = USB_MSC_MEM_SIZE;
  configASSERT(requiredSize <= constSize);

  initParam.mem_base = USB_MSC_MEM_BASE;
  initParam.mem_size = constSize;
  
  initParam.MSC_Write = MSCWrite;
  initParam.MSC_Read = MSCRead;
  initParam.MSC_Verify = MSCVerify;
  
  initResult = mUsbApi->msc->init(mUsbHandle, &initParam);
  configASSERT(initResult == LPC_OK);
}

static void InitializeMicroSd(void)
{
  DRESULT initResult;
  initResult = disk_initialize(0);
  configASSERT(initResult == RES_OK);
}

static DWORD GetSectorCount(void)
{
  DRESULT result;
  DWORD count;
  result = disk_ioctl(0, GET_SECTOR_COUNT, &count);
  configASSERT(result == RES_OK);
  return count;
}


void UsbMscMain(void *param)
{
  uint32_t sectorCount;
  
  mUsbApi = (USBD_API_T*)((*(ROM **)(0x1FFF1FF8))->pUSBD);

  // MicroCard SDを初期化
  InitializeMicroSd();
  sectorCount = GetSectorCount();
  
  // GPIO / USB / USBRAM
  LPC_SYSCON->SYSAHBCLKCTRL |= (1 << 6) | (1 << 14) | (1 << 27);
  CLEAR_SET_BIT32(LPC_IOCON->PIO0_3, 0x07, 0x01); // USB_VBUS
  CLEAR_SET_BIT32(LPC_IOCON->PIO0_6, 0x07, 0x01); // USB_CONNECT
  
  InitializeUsbCore();
  InitializeUsbMsc(sectorCount);

  NVIC_SetPriority(USB_IRQn, 0);
  NVIC_EnableIRQ(USB_IRQn);
  mUsbApi->hw->Connect(mUsbHandle, 1);
  
  while (1) {
    // USB接続が切れていないかポーリングで確認し続ける
    if (!(LPC_USB->DEVCMDSTAT & (1 << 28))) {
      // 接続が切れたらLPCをリセットする
      vTaskDelay(500);
      NVIC_SystemReset();
    }
    vTaskDelay(50);
  }
}
