﻿// -*- Encoding: UTF-8 -*-
#include "FreeRTOSCommon.h"
#include "LPC11Uxx.h"
#include "GPS.h"
#include "ff.h"
#include "diskio.h"
#include "UsbMsc.h"
#include "Sound.h"
#include "Speech.h"
#include "RomSounds.h"
#include "ConstMem.h"
#include "bitop.h"
#include "LED.h"
#include "PitchMeasure.h"
#include "PitchAdjust.h"
#include "DestinationMgr.h"
#include <stdio.h>
#include <stdlib.h>

#define GPS_UPDATE_INTERVA_SEC  30
// #define ENABLE_APPEND_LOG       1

void EnterDeepPowerDown(void)
{
  LPC_PMU->PCON = 0x03;
  SCB->SCR |= SCB_SCR_SLEEPDEEP_Msk;
  __WFI();
  NVIC_SystemReset();
}

void vApplicationTickHook( void )
{
  disk_timerproc();
}

void vApplicationMallocFailedHook (void)
{
  DEBUG_MSG(("[ERR] Failed Malloc\n"));
  while (1);
}

void vApplicationStackOverflowHook( xTaskHandle *pxTask, signed portCHAR *pcTaskName )
{
  DEBUG_MSG(("[ERR] StackOver\n"));
  while (1);
}
     
DWORD get_fattime (void)
{
  GPS_TIME tm;
  GetGPSTime(&tm);
  return ( ((((DWORD)tm.yearBase2000 + 20) & 0x7F) << 25)
           | ((DWORD)tm.mon << 21)
           | ((DWORD)tm.day << 16)
           | ((DWORD)tm.hour << 11)
           | ((DWORD)tm.min << 5)
           | ((DWORD)tm.sec >> 1));
}

static void PowerOnPeripheral(void)
{
  LPC_GPIO->SET[1] |= (1 << 27);
}

static void PowerOffPeripheral(void)
{
  LPC_GPIO->CLR[1] |= (1 << 27);
}

bool IsPushSwitchDown(void)
{
  if (LPC_GPIO->B0[16]) {
    return false;
  }
  return true;
}

static void StepTimerCallback(xTimerHandle timer)
{
  PlaySoundInROM(gPiSound, sizeof(gPiSound), 0);
}

static bool IsUsbConnected(void)
{
#if DEBUG
  // 評価中はつねにfalse
  return false;
#endif
  
  if (LPC_GPIO->B0[3]) {
    return true;
  }
  return false;
}

static void AnnounceGPSPitchState(bool pitchValid)
{
  static uint32_t prevTick = 0;
  uint32_t currentTick = xTaskGetTickCount();
  if ((currentTick - prevTick) > 3000) {
    if (!IsGPSActive()) {
      PlaySoundInSD("m/search_g.wav", portMAX_DELAY);
    } else if (!pitchValid) {
      PlaySoundInSD("m/pitch_m.wav", portMAX_DELAY);
    } else {
      // ピッチ音を出しているので発声不要
    }
    prevTick = xTaskGetTickCount();
  }
}

static void AnnounceEnd(const char* fname)
{
  // 同じメッセージを繰り返す
  while (!IsPushSwitchDown()) {
    PlaySoundInSD(fname, portMAX_DELAY);
    vTaskDelay(500);
  }
}

static void AnnounceEstimatedArrivalTime(int diffArrivalTime)
{
  if (diffArrivalTime == 0) {
    return;
  }
  PlaySoundInSD("m/est.wav", portMAX_DELAY);
  SpeakPlusMinus(diffArrivalTime);
  SpeakTimestamp(abs(diffArrivalTime));
}

static int ConvertStepIntervalToBpm(int stepInterval)
{
  return (60 * 1000) / stepInterval;
}

#ifdef ENABLE_APPEND_LOG
static void AppendLog(uint32_t StepInterval,
                      uint32_t prevPreferredStepInterval,
                      uint32_t newPreferredStepInterval,
                      int diffArrivalTime,                      
                      GPS_STEP_INFO *prev, GPS_STEP_INFO *new)
{
  FRESULT fr;
  UINT bw;
  FIL *fp = gpFil;
  fr = f_open(fp, "justlog.bin", FA_WRITE | FA_READ | FA_OPEN_ALWAYS);
  if (fr != FR_OK) {
    return;
  }
  fr = f_lseek(fp,f_size(fp));
  if (fr != FR_OK) {
    f_close(fp);
    return;
  }
  f_write(fp, &StepInterval, 4, &bw);
  f_write(fp, &prevPreferredStepInterval, 4, &bw);
  f_write(fp, &newPreferredStepInterval, 4, &bw);
  f_write(fp, &diffArrivalTime, 4, &bw);
  f_write(fp, prev, sizeof(GPS_STEP_INFO), &bw);
  f_write(fp, new, sizeof(GPS_STEP_INFO), &bw);
  f_close(fp);
}
#endif

static int UpdateStepInterval(int currentPreferredStepInterval,
                              GPS_STEP_INFO *prev, GPS_STEP_INFO *new)
{
  int diffArrivalTime;
  
  uint32_t currentStepInterval = CalcStepInterval(prev, new);
  uint32_t newPreferredStepInterval = CalcPreferedStepInterval(currentStepInterval, prev, new, &diffArrivalTime);

#ifdef ENABLE_APPEND_LOG
  AppendLog(currentStepInterval,
            currentPreferredStepInterval,
            newPreferredStepInterval,
            diffArrivalTime,
            prev, new);
#endif

  AnnounceEstimatedArrivalTime(diffArrivalTime);
  vTaskDelay(1000);
  
  if (currentPreferredStepInterval != 0) {
    if (currentPreferredStepInterval != newPreferredStepInterval) {
      // 今歩いているピッチからの修正がある場合は、発声
      if (newPreferredStepInterval < currentStepInterval) {
        PlaySoundInSD("m/up.wav", portMAX_DELAY); // 上方修正
      } else if (newPreferredStepInterval > currentStepInterval) {
        PlaySoundInSD("m/down.wav", portMAX_DELAY); // 下方修正
      }
    }
  }
  if (currentPreferredStepInterval != newPreferredStepInterval) {
    SpeakBpm(ConvertStepIntervalToBpm(newPreferredStepInterval));
  }
  return newPreferredStepInterval;
}

static bool EndCheck(xTimerHandle timer)
{
  double lat,lng;
  
  // 目的地周辺に到着している場合は終了メッセージ
  GetGPSLatLng(&lat, &lng);
  if (IsArrived(lat, lng)) {
    xTimerStop(timer, 0);
    AnnounceEnd("m/arrived.wav");
    return true;
  }

  // すでに到着時刻を過ぎてしまっている
  if (GetGPSTimestamp() >= GetFinalScheduledArrivalTime()) {
    xTimerStop(timer, 0);
    AnnounceEnd("m/damepo.wav");
    return true;
  }

  return false;
}

static void MainLoop(xQueueHandle eventQueue)
{
  xTimerHandle stepTimer;
  uint8_t event;
  bool destinatinLoaded = false;
  uint32_t preferredStepInterval = 0;
  uint32_t newStepInterval = 0;
  GPS_STEP_INFO prevInfo = {0};
  GPS_STEP_INFO newInfo = {0};
  
  stepTimer = xTimerCreate((const signed char*)"PIT",
                            600, // 仮の値
                            pdTRUE, 0, StepTimerCallback);
  
  while (1) {
    if (xQueueReceive(eventQueue, &event, 1000)) {
      switch (event) {
      case GPS_EVENT_ACTIVE:
        DEBUG_MSG(("GPS ACTIVE\n"));
        if (!destinatinLoaded) {
          if (!LoadDestinationInfo(GetGPSTimestamp())) {
            AnnounceEnd("m/dsterr.wav");
            return;
          }
          destinatinLoaded = true;
        }
        LedOn();
        if (IsWalking()) {
          ChangeGPSUpdateInterval(GPS_UPDATE_INTERVA_SEC);
        }
        break;
        
      case GPS_EVENT_INACTIVE:
        DEBUG_MSG(("GPS INACTIVE\n"));
        LedBlink(500);
        prevInfo.valid = 0;
        ChangeGPSUpdateInterval(0);
        break;
        
      case GPS_EVENT_UPDATE:
        DEBUG_MSG(("GPS UPDATE\n"));

        UpdateGPSStepInfo(&newInfo);
        if (newInfo.valid & LAT_LNG_INFO_VALID) {
          if (UpdateNextDestination(newInfo.lat, newInfo.lng)) {
            PlaySoundInSD("m/point.wav", portMAX_DELAY);
          }
        }
        
        if (IsNeededInfoValid(&newInfo) && IsNeededInfoValid(&prevInfo)) {
          xTimerStop(stepTimer, 0);
          
          // ピッチ更新 & アナウンス
          newStepInterval = UpdateStepInterval(preferredStepInterval, &prevInfo, &newInfo);
          if (preferredStepInterval != newStepInterval) {
            xTimerChangePeriod(stepTimer, newStepInterval, 0);
            preferredStepInterval = newStepInterval;
          }

          xTimerStart(stepTimer, 0);
        }
        
        prevInfo = newInfo;
        break;
        
      case STEP_EVENT_STOP:
        DEBUG_MSG(("STEP STOP\n"));
        
        // 歩みが止まったので、情報無効化 GPS Updateイベント停止
        prevInfo.valid = 0;
        ChangeGPSUpdateInterval(0);
        xTimerStop(stepTimer, 0);
        break;
        
      case STEP_EVENT_WALK:
        DEBUG_MSG(("STEP WALK\n"));
        
        // 歩き始めたので現時点の状態を保存してGPS Updateイベント開始
        UpdateGPSStepInfo(&prevInfo);
        if (IsGPSActive()) {
          ChangeGPSUpdateInterval(GPS_UPDATE_INTERVA_SEC);
        }
        if (preferredStepInterval != 0) {
          xTimerStart(stepTimer, 0);
        }
        break;
        
      default:
        break;
      }
    } else {

      if (destinatinLoaded) {
        if (IsGPSActive()) {
          double lat, lng;
          GetGPSLatLng(&lat, &lng);
          if (UpdateNextDestination(lat, lng)) {
            PlaySoundInSD("m/point.wav", portMAX_DELAY);
          }
        }
        if (EndCheck(stepTimer)) {
          return;
        }
      }
      
      if (IsPushSwitchDown()) {
        // 終了
        return;
      }
      
      // USBが接続されたらリセットする
      if (IsUsbConnected()) {
        vTaskDelay(500);
        NVIC_SystemReset();
      }
      
      AnnounceGPSPitchState((preferredStepInterval != 0) ? true : false);
    }
  }
}

static void MainTask(void *param)
{
  xQueueHandle eventQueue;
  
  gpFatFs = pvPortMalloc(sizeof(FATFS));
  gpFil = pvPortMalloc(sizeof(FIL));
  
#ifndef DEBUG
  // デバッグビルドではPowerOn時のDeep Power Down処理はスキップ
  vTaskDelay(1000);
  if (!IsPushSwitchDown()) {
    EnterDeepPowerDown();
    // Never Return
  }
#endif

  eventQueue = xQueueCreate(16, 1);

  PowerOnPeripheral();
  vTaskDelay(200);
  
  InitializeLED();
  LedBlink(200);

  f_mount(gpFatFs, "", 0);
  InitializeSound();
  InitializePitchMeasure(eventQueue);
  InitializeGPS(eventQueue);

  LedBlink(500);
  
  while (IsPushSwitchDown()) {
    vTaskDelay(100);
  }
  
  MainLoop(eventQueue);
  
  LedOff();
  vTaskDelay(1000);
  while (IsPushSwitchDown()) {
    vTaskDelay(100);
  }
  PowerOffPeripheral();
  EnterDeepPowerDown();
  NVIC_SystemReset();
}

int main(void)
{
  LPC_SYSCON->SYSAHBCLKCTRL |= (1 << 6);            // GPIO有効化
  InitializeConstMem();
  
  CLEAR_SET_BIT32(LPC_IOCON->PIO1_27, 0x07, 0x00);  // P1.27 GPIO
  LPC_GPIO->DIR[1] |= (1 << 27);                    // P1.27 OUTPUT
  PowerOffPeripheral();
  
  if (IsUsbConnected()) {
    PowerOnPeripheral();
    SetupUsbModeHeap();
    LedOnWithoutOS();
    xTaskCreate(UsbMscMain, (const signed char*)"USB", 150, NULL, APP_NORMAL_PRIORITY, NULL);
  } else {
    SetupNormalModeHeap();
    xTaskCreate(MainTask, (const signed char*)"Main", 250, NULL, APP_NORMAL_PRIORITY, NULL);
  }
  vTaskStartScheduler();
  while(1);
  return 0;
}
